!
!******************************************************************
! THIS PROGRAM PERMITS TESTING THE OBSERVATIONAL OPERATOR ROUTINE 
!  
!******************************************************************
! VERSION: 1.1
! WRITTEN BY: NOVELTIS (A.KLONECKI)
! THE PROGRAM IS BASED ON THE ALGORITHMS PRESENT IN THE ASSIMILATION CODE
! WRITTEN BY JF Lamarque 
! 
! CREATION DATE: 22/11/2005
! REVISION DATE: 12/01/2006
!    -generalized the program, now can handle also MIPAS data, as well as data
!    with varying number of vertical levels
!    -added new flags to handle the different cases: vertical integration,
!    data with and without averaging kernel, temporal interpolation
!    -added new inputs that filter the observed data (latmin,latmax,presmin,
!    presmax,dates.
!    -changed the content of the data structure with observations (in 
!    (data_structures.f90): added pressure, removed variables which are 
!    not needed for the observational operator.
!    -cleaned up the main program and the routine observ_operator.f90
!
! REVISION DATE: 13/07/2006 
!   -modified the routines : multiply_matrix_csr_vector, and multiply_matrices_csr. Now 
!    there are no dynamical allocations in the routines for variables that are in the 
!    input/output list. 
!   -added a few: deallocate to prevent memory loses.
!
!******************************************************************
! the inputs from the model (latitude,longitude,CO, etc) are specified 
! in the program. They can be also read from files with model output-
! an appropriate routine needs to be inserted. 
!
! Similarly, the averaging functions are set to constant values in the
! program. More realistic values must be read from separate files. 

! OUTLINE OF THE PROGRAM:
! 1) set the integrated averaging kernel to constant values 
! 2) set the model variables
! 3) read the observation data (IASI CO columns)
! 4) call the observational operator which computes H
! 5) compute H*Xmod
! 6) compute H*Xmod+(I-A)*Xa if needed
 
!******************************************************************

  program test_observ_operator 

  use floats
  use transformations_matrix
  use matrix_manipulations
  use charger_obs_data
  use MODEL_STRUCTURE
  use MATRIX_CSR_STRUCTURE
  use OBS_STRUCTURE


  implicit none 
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! DECLARATIONS:
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!    Dimensioning variables:
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!maximum numb. of obs veritcal levels:
     integer*4                            :: n_lvls     
!number of vertical levels for the output variable. If column data: 1
!      if profiles : n_lvls.
     integer*4                            :: n_lvls_out 
!numb. of model longitudes and latitudes                           
     integer*4                            :: nlon,nlat  
!numb. of model vertical levels
     integer*4                            :: nlev     
!numb. of temporary MOCAGE files: if 1 no temporal interpolation
!                                 if 2 - temporal interpolation  
     integer*4  :: ntime      
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! OBS DATA: 
! the variable used for storing the averaging kernel (n_lvls_out,n_lvls)
     real(std),pointer,dimension(:,:) :: avg_kernel
! apriori contribution (n_lvls):
     real(std),pointer,dimension(:)   :: apriori_contrib
! surface pressure in the model, for now constant
     real                                 :: ps
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! file names
      character*200                       :: Data_dir
      character*200                       :: Data_file
      character*200                       :: MOCAGE_DIR
      character*200                       :: Kernel_Dir
      character*200                       :: Atm_Dir
      character*200                       :: filename
      character*9                         :: OBS_DATA_TYPE
!
      character*4                         :: char4
!
! INPUT FLAGS:
      logical                             :: flag_integrated_column
      logical                             :: flag_temp_interp
      logical                             :: flag_averaging_kernel
! Variables used to filter the observed data:
      real(std)                           :: latmin,latmax
      real(std)                           :: lonmin,lonmax
      real(std)                           :: presmin,presmax
      integer,dimension(2)                :: MOCAGE_day
      integer,dimension(2)                :: MOCAGE_hour
      real(std),   dimension(2)           :: dates_MOCAGE
! Structures with the modeled and observed data:
      type(MODEL_DATA)                    :: Mod_data
      type(OBSERV_DATA)                   :: Obs_data

! variable that contains info on the obs for which the matrix H could
! not be calculated (because the obs is not in the model domain).
      logical,pointer,dimension(:)    :: discard_obs
! The final matrix=observational operator:
      type(CSR_FORMAT)                    :: Hres_csr
! The final result containing the model data in observational space.
      real(std),pointer,dimension(:,:):: traceur_inobs_space
! function:
      integer                             :: ltrim
! indexing variables:
      integer                             :: i,j,k,it
!******************************************************************
! END OF DECLARATIONS
!******************************************************************

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! the code starts here
! read the input variables:
      read(*,*,err=999) Data_dir
      read(*,*,err=999) OBS_DATA_TYPE           !type of observed data
      read(*,*,err=999) flag_temp_interp        !perform temporal interp ?
      read(*,*,err=999) flag_averaging_kernel   !apply averaging kernel ? 
      read(*,*,err=999) flag_integrated_column  !integrate the obs column ? 
      read(*,*,err=999) latmin
      read(*,*,err=999) latmax
      read(*,*,err=999) lonmin
      read(*,*,err=999) lonmax

      if (flag_temp_interp) then 
        write(*,*) 'DOING THE TEMPORAL INTERPOLATION'
        ntime=2
      else
        write(*,*) 'NO TEMPORAL INTERPOLATION'
        ntime=1
      endif

      if (flag_integrated_column) then
        write(*,*) 'assimilating column data'
        if (.not.flag_averaging_kernel) then 
          write(*,*) 'need to set the averaging kernel for column data'
          write(*,*) 'Exiting...'
          call exit(-1)
        end if
      else
       write(*,*) 'assimilating observed profiles'
      end if


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
     
     

      if (OBS_DATA_TYPE.eq.'IASI_data') then 
        MOCAGE_day(1)= 1652
        MOCAGE_hour(1)= 6
        MOCAGE_day(2)= 1652
        MOCAGE_hour(2)= 12
        presmin=2500
        presmax=100000
        n_lvls=20 
      else if (OBS_DATA_TYPE.eq.'MIPAS_dat') then
        MOCAGE_day(1)=1277
        MOCAGE_hour(1)= 6
        MOCAGE_day(2)= 1277
        MOCAGE_hour(2)= 12
        presmin=2500
        presmax=100000
        n_lvls=18
      else
         write(*,*) 'INCORRECT TYPE OF OBSERVED DATA, EXITING'
         call exit(-1)
      end if
! 1) READ THE AVERAGING KERNEL INFO
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
     if (flag_integrated_column) then
       n_lvls_out=1
     else
       n_lvls_out=n_lvls
     end if 
  
     if ( flag_averaging_kernel ) then 
       allocate(avg_kernel(n_lvls_out,n_lvls))
     allocate(apriori_contrib(n_lvls_out))

       if (flag_integrated_column) then 
         apriori_contrib(:)=-4.e+15
         do i=1,n_lvls
           avg_kernel(:,i)=1.e23
         end do
       else
         apriori_contrib(:)=10.E-9
         do i=1,n_lvls
           avg_kernel(:,i)=0.1
         end do
       end if
     else
! need to allocate, but set to 0, will not be used later
       allocate(avg_kernel(n_lvls_out,n_lvls))
       avg_kernel=0.
     end if
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! 2) Assign THE MOCAGE data (CO,PRESSION,latitude,longitude, as well as the 
!    dimensions: nlat,nlon,nlev_m)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! MOCAGE dates:
      do i=1,ntime
        dates_MOCAGE(i)=MOCAGE_day(i)+float(MOCAGE_hour(i))/24.d0
      end do
      
      nlat=90
      nlon=180
      nlev=40
      Mod_data%nlat = nlat
      Mod_data%nlon = nlon
      Mod_data%nlev = nlev

      allocate(Mod_data%traceur(nlon,nlat,nlev,ntime))
      allocate(Mod_data%pres(nlon,nlat,nlev,ntime))
      allocate(Mod_data%lat(nlat,nlon))
      allocate(Mod_data%lon(nlat,nlon))

      ps = 100000.
      call random_number(Mod_data%traceur)
      do it=1,2
        do i=1,Mod_data%nlon
          do j=1,Mod_data%nlat
            Mod_data%pres(i,j,Mod_data%nlev,it)=ps
            do k=Mod_data%nlev-1,1,-1
              Mod_data%pres(i,j,k,it)=Mod_data%pres(i,j,k+1,it)-2500.
            end do
            do k=1,Mod_data%nlev
              Mod_data%traceur(i,j,k,it)=       &
                (200.+Mod_data%traceur(i,j,k,it)*25)*1.e-9
            end do
          end do
        end do
      end do

      do j=1,Mod_data%nlon
      do i=1,Mod_data%nlat
        Mod_data%lon(i,j) = (        2. * (j-1)) 
      end do
      end do
      do j=1,Mod_data%nlon
      do i=1,Mod_data%nlat
        Mod_data%lat(i,j) = (-89 + 2. * (i-1)) 
      end do
      end do

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! 3) reads the OBS data
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      if (OBS_DATA_TYPE.eq.'IASI_data') then 
        write(char4,'(i4.4)') MOCAGE_day(1)
        Data_file=Data_dir(1:ltrim(Data_dir))//'IASI.L2.'//char4//'.nc'
        write(*,*) 'IASI L2 data are taken from: ',Data_file(1:ltrim(Data_file))
        filename=Data_file

        call read_IASI_netcdf(dates_MOCAGE,filename      &
                         ,n_lvls          &
                         ,latmin          &
                         ,latmax          &
                         ,lonmin          &
                         ,lonmax          &
                         ,presmin         &
                         ,presmax         &
                         ,Obs_data        &
                          )
        write(*,*) 'READ IASI L2 CO DATA FOR ',Obs_data%n_prf,     &
                   ' IASI CO integrated columns'
       else if (OBS_DATA_TYPE.eq.'MIPAS_dat') then 
         write(*,*) 'reading the MIPAS data'
         Data_file=Data_dir(1:ltrim(Data_dir))//'HOBS20030701.txt'
         filename=Data_file
         call read_mipas_data(dates_MOCAGE,filename      &
                         ,n_lvls          &
                         ,latmin          &
                         ,latmax          &
                         ,lonmin          &
                         ,lonmax          &
                         ,presmin         &
                         ,presmax         &
                         ,Obs_data        &
                          )

       end if
! end reading OBS data

      allocate(discard_obs(Obs_data%n_prf))

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! 4) compute the observational operator
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      call observ_operator(                 &
! INPUT MODEL DATA:
                            Mod_data%nlat        & !dimensions:
                           ,Mod_data%nlon        &
                           ,Mod_data%nlev        &
                           ,ntime                &
                           ,Mod_data%lat         & 
                           ,Mod_data%lon         & 
                           ,Mod_data%pres        &
                           ,dates_MOCAGE         &
! INPUT OBS DATA:
                           ,Obs_data%n_prf       & 
                           ,n_lvls               &
                           ,n_lvls_out           &
                           ,Obs_data%lat         & 
                           ,Obs_data%lon         &
                           ,Obs_data%date        &
                           ,Obs_data%pres        &
! INPUT COMPLEMENTARY DATA:
                           ,avg_kernel           &
! EXECUTION OPTIONS:
                           ,flag_averaging_kernel &
                           ,flag_temp_interp     &
! OUTPUT DATA!
                           ,Hres_csr             & 
                           ,discard_obs          &
                           )
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! 5) compute H*Xmod 
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
     allocate(traceur_inobs_space(n_lvls_out,Obs_data%n_prf))
     call multiply_observ_operator(Hres_csr,Mod_data,ntime,                &
                                   discard_obs,Obs_data%n_prf,n_lvls_out,  &
                                   traceur_inobs_space)
     call deallocate_operator(Hres_csr)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! 6) compute H*Xmod+(I-A)*Xa if needed
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
     if (flag_averaging_kernel) then 
       do i=1,Obs_data%n_prf
         if (.not.discard_obs(i)) then
           traceur_inobs_space(:,i)=traceur_inobs_space(:,i)+apriori_contrib(:)
         end if
       end do
     end if
     open(10,file='output.dat')
     do i=1,Obs_data%n_prf
       write(10,100) i,(traceur_inobs_space(j,i),j=1,n_lvls_out)
100    format(i5,50(1PE14.5))
     end do
     close(10)

     deallocate(traceur_inobs_space)

     deallocate(discard_obs)

     write(*,*) 'SUCCESSFUL TERMINATION OF THE PROGRAM'
     stop

999  write(*,*) 'error reading the input parameters, exiting'
     call exit(-1)

     end 
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
