      module transformations_matrix
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!this module contains the subroutines used in the transformation of the MOCAGE
! data from model space to observational space. The routines return the 
! interpolation matrices corresponding to the interpolation matrices. The matrices
! are represented in the Compressed Sparse Row (CSR) format to avoid creating
! and manipulating the matrices with full model dimensions.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! CSR FORMAT:
! IN THE CSR FORMAT EACH MATRIX IS REPRESENTED BY THREE ARRAYS (see the document 
! contained in the CSR documentation): 
! -THE REAL ARRAY CONTAINING THE NON-ZERO ELEMENTS (SIZE NNZ)
! -INTEGER ARRAY (JZ) CONTAINING THE COLUMN INDEX FOR EACH NON-ZERO ELEMENT 
!  (SIZE NNZ) 
! -INTEGER ARRAY (IZ) CONTAINING THE POINTERS (POSITIONS) TO THE ELEMENTS IN 
!  THE REAL ARRAY. THE DIMENSIONS ARE NLEV+1, WITH THE FIRST NLEV ELEMENTS POINTING 
!  TO THE LOCATION OF THE FIRST ELEMENT IN EACH OF THE ROWS OF THE FULL MATRIX. THE 
!  ELEMENT N+1 CONTAINS : IA(1)+NNZ  
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
 
! 
! VERSION: 1.0
! WRITTEN BY: NOVELTIS (A.KLONECKI)
! CREATION DATE: 26/09/2005
! REVISION DATE:

contains 

!
!******************************************************************
! THE ROUTINE transf_horiz_interp_matrix IS RESPONSIBLE FOR PREPARING THE MATRIX
! FOR THE HORIZONTAL INTERPOLATION OF THE MODEL DATA TOWARDS THE HORIZONTAL GRID 
! OF THE OBSERVATIONS. THE INPUT DATA IS GIVEN FOR ALL MODEL LEVELS AND FOR ONE
! OR TWO MOCAGE TIME PERIODS.THE OUTPUT MATRIX IS GIVEN IN THE CRS FORMAT
!******************************************************************
!
      subroutine transf_horiz_interp_matrix(NPRF,nlon,nlat,nlev,ntime,        &
                                            latgrid,longrid,lat_o,lon_o,          &
                                            discard_obs,                      &
                                            H_csr)
      use floats
      use MATRIX_CSR_STRUCTURE
      use matrix_manipulations

      implicit none

! INPUT VARIABLES:

      integer  ,intent(in)  :: NPRF       !numb. of obser. profiles present in 
                                          !the current MOCAGE interval
      integer*4,intent(in)  :: nlon,nlat  !numb. of model longitudes and latitudes 
      integer*4,intent(in)  :: nlev       !numb. of model vertical levels          
      integer*4,intent(in)  :: ntime      !numb. of model outputs (1 if no temporal 
                                          !interpolation, 2 if temp. interpolation 
                                          !between two model outputs.
! Model lats and longs
      real(std), dimension(nlat,nlon),intent(in)         :: latgrid
      real(std), dimension(nlat,nlon),intent(in)         :: longrid
! IASI lats and longs
      real(std),dimension(NPRF),intent(in)                 :: lat_o
      real(std),dimension(NPRF),intent(in)                 :: lon_o   

! OUTPUT VARIABLES:
!  Obs that are not in the model domain
      logical,dimension(NPRF),intent(out)             :: discard_obs
!  The horizontal interp matrix in the CSR representation (structure with 3 arrays):  
      type(CSR_FORMAT)                                :: H_csr

! LOCAL VARIABLES:
!  variables used for horizontal interpolations:       
!  the 4 model pixels encircling the obs (NPRF,4) : 
      integer,   pointer         :: nhxy(:,:)       
!  weights for the 4 pixels, used in horiz. interpol. (NPRF,4) :
      real(std),    pointer         :: hxy(:,:)
!
!  variables used for matrix representation:
      integer*4 nmax 
      integer*4 i1,i2,j1,j2
      real(std)    wj1,wj2,wi1,wi2
!  indexing variables:
      integer*4 :: i,n,k,it
      integer*4 :: icounter
      integer*4 :: ind1,ind2,ind3,ind4,indk
!
      integer*4 :: ALLOC_ERR
      integer*4,parameter  :: NUMB_NON_ZERO=4  !number of non-zero elements
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! the code stars here:
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
      allocate(hxy (NPRF,4),STAT=ALLOC_ERR)
      call check_allocate(ALLOC_ERR,"transf_horiz_interp_matrix")
      allocate(nhxy(NPRF,4),STAT=ALLOC_ERR)
      call check_allocate(ALLOC_ERR,"transf_horiz_interp_matrix")

! calculate the weights:
      call define_hxy(nlon,nlat,NPRF        & 
                   ,latgrid,longrid         &  
                   ,lat_o                     &  
                   ,lon_o                     &  
                   ,hxy,nhxy,discard_obs)
!      print *,'define_hxy called'
!
! define the dimensions of the CRS arrays
      H_csr%NPRF  = NPRF
      H_csr%nrow  = nlev*ntime
      H_csr%ncol  = nlat*nlon*nlev*ntime 
      nmax        = NUMB_NON_ZERO*H_csr%nrow  !there are 4 non 0 elements in each row
      H_csr%nmax  = nmax

      allocate(H_csr%H(nmax,NPRF),STAT=ALLOC_ERR)
      call check_allocate(ALLOC_ERR,"transf_horiz_interp_matrix")
      allocate(H_csr%j(nmax,NPRF),STAT=ALLOC_ERR)
      call check_allocate(ALLOC_ERR,"transf_horiz_interp_matrix")
      allocate(H_csr%i(H_csr%nrow+1,NPRF),STAT=ALLOC_ERR)
      call check_allocate(ALLOC_ERR,"transf_horiz_interp_matrix")


      do n=1,NPRF
        if (discard_obs(n)) then 
          write(*,*) 'WARNING, found observations that are not in the model domain!'
          write(*,*) 'observation number ',n,'lat,lon',lat_o(n),lon_o(n)
          H_csr%H(:,n)=0.
          H_csr%i(:,n)=0
          H_csr%j(:,n)=0
        else
! construct the matrix for the horizontal interpolation in the 
! CSR format          
!
! positions of the 4 model pixels:
          j1  = nhxy(n,1)   !lat1
          j2  = nhxy(n,2)   !lat2
          i1  = nhxy(n,3)   !lon1
          i2  = nhxy(n,4)   !lon2
    
          IF(i1.lt.0.or.i2.lt.0.or.j1.lt.0.or.j2.lt.0) CYCLE

! weights for the 4 model pixels (in the j direction, and in i direction):
          wj1 = hxy (n,1)
          wj2 = hxy (n,2)
          wi1 = hxy (n,3)
          wi2 = hxy (n,4)
!
          ind1=(j1-1)*nlon+i1
          ind2=(j1-1)*nlon+i2
          ind3=(j2-1)*nlon+i1
          ind4=(j2-1)*nlon+i2

! construct the operator matrix: (for explanation see the 
! corresponding document)
! the dimensions of the full matrix are : 
! Rows=ntime*nlev ,  Columns=nlat*nlon*nlev*ntime  
! 
          icounter=1
          do it=1,ntime
            do k=1,nlev
! assign the real array with non-zero weights
              H_csr%H(icounter,  n)=wi1*wj1
              H_csr%H(icounter+1,n)=wj1*wi2
              H_csr%H(icounter+2,n)=wj2*wi1
              H_csr%H(icounter+3,n)=wj2*wi2
!assign the integer array with the column positions for non-zero weights
              indk=(it-1)*nlat*nlon*nlev+(k-1)*nlat*nlon 
              H_csr%j(icounter  ,n)=indk+ind1
              H_csr%j(icounter+1,n)=indk+ind2
              H_csr%j(icounter+2,n)=indk+ind3
              H_csr%j(icounter+3,n)=indk+ind4
!
              icounter=icounter+NUMB_NON_ZERO
!
            end do
!assign the integer array with the positions of the first non-zero element
! in each row of the full matrix
            do i=1,nlev
              H_csr%i((it-1)*nlev+i,n)=(it-1)*nlev*NUMB_NON_ZERO+  &
                            (i-1)*NUMB_NON_ZERO+1
            end do
          end do
          H_csr%i(ntime*nlev+1,n)=H_csr%i(1,n)+H_csr%nmax

! matrix construction finished
!
        end if !if discard_obs
      end do  !n=1,NPRF
!
      deallocate(hxy)
      deallocate(nhxy)

      return
      end subroutine transf_horiz_interp_matrix

!******************************************************************
!******************************************************************
! THE ROUTINE transf_vert_interp_matrix IS RESPONSIBLE FOR THE PREPARING THE 
! MATRIX CORRESPONDING TO THE VERTICAL INTERPOLATION OF THE 
! MODEL DATA TOWARDS THE VERTICAL (pressure) GRID OF THE OBSERVATIONS. 
! THE INPUT DATA IS GIVEN FOR ONE or TWO MOCAGE TIME PERIODS.
!******************************************************************
!******************************************************************
!******************************************************************
      subroutine transf_vert_interp_matrix(pres_inth,                    &
                                           pres_o,                       &
                                           discard_obs,                  &
                                           NPRF,nlev,mlev,ntime,         &
                                           H_csr)
 
      use floats 
      use MathLib
      use MATRIX_CSR_STRUCTURE
      use matrix_manipulations

      implicit none 

! INPUT VARIABLES
      integer*4,intent(in)       :: nlev  !model levels
      integer*4,intent(in)       :: mlev  !obs levels
      integer*4,intent(in)       :: NPRF
      integer*4,intent(in)       :: ntime

      real(std), dimension(ntime*nlev,NPRF),intent(in)     :: pres_inth   !model pressures
      real(std), dimension(mlev,NPRF)      ,intent(in)     :: pres_o      !obs pressures
      logical,dimension(NPRF)              ,intent(in)     :: discard_obs !obs to discard
! 
! OUTPUT VARIABLES
!  The vertical interpolation matrix in the CSR representation (structure with 3 arrays):
      type(CSR_FORMAT)    :: H_csr  !REAL ARRAY

!  information on the size of the arrays in the CSR format:

!
! LOCAL VARIABLES
! VARIABLES FOR THE CONSTRUCTION OF THE MATRIX
      integer                          :: nmax
!
      real(std), dimension(nlev)       :: unit                !unit vector
      integer*4                        :: ind1,ind2,ind3,ind4 !indices
      real(std),dimension(mlev)        :: temp                !temporary array for storing
                                                           !the contribution of a model level
                                                           !to all observed levels

      integer*4                  :: k
      integer*4                  :: ip
      integer*4                  :: it
      integer*4                  :: i_nlev,i_mlev
      integer*4                  :: ierr
      integer*4 :: ALLOC_ERR

      real(std), pointer    ,dimension(:)   :: weights
      integer,pointer    ,dimension(:)   :: positions

      integer*4,parameter  :: NUMB_NON_ZERO=2  !number of non-zero elements in each row
! construct the matrix:
!
! define the dimensions of the CRS arrays
      H_csr%NPRF  = NPRF
      H_csr%nrow  = mlev*ntime
      H_csr%ncol  = nlev*ntime
      nmax        = NUMB_NON_ZERO*H_csr%nrow      !there are 2 non 0 elements in each row
      H_csr%nmax=nmax
!
      allocate(H_csr%H(nmax,NPRF),STAT=ALLOC_ERR)
      call check_allocate(ALLOC_ERR,"transf_vert_interp_matrix")
      allocate(H_csr%j(nmax,NPRF),STAT=ALLOC_ERR)
      call check_allocate(ALLOC_ERR,"transf_vert_interp_matrix")
      allocate(H_csr%i(H_csr%nrow+1,NPRF),STAT=ALLOC_ERR)
      call check_allocate(ALLOC_ERR,"transf_vert_interp_matrix")
!      
      allocate(weights(nmax),STAT=ALLOC_ERR)
      call check_allocate(ALLOC_ERR,"transf_vert_interp_matrix")
      allocate(positions(nmax),STAT=ALLOC_ERR)
      call check_allocate(ALLOC_ERR,"transf_vert_interp_matrix")
!
      do ip=1,NPRF
        if (.not.discard_obs(ip)) then 
          do it=1,ntime
            ind1=(it-1)*mlev*NUMB_NON_ZERO
            ind2=it*mlev*NUMB_NON_ZERO
            ind3=(it-1)*nlev
            ind4=it*nlev
! get the weights and the indexes of the model levels used for interpolating
! to pressure levels:
            call get_coef_interpol(pres_inth(ind3+1:ind4,ip),     &
                  pres_o(:,ip),                                   &
                  weights(ind1+1:ind2),positions(ind1+1:ind2))
          end do
!construct the matrix:
! fill the upper half of the matrix:
          do k=1,H_csr%nrow
            H_csr%H(k,ip)=weights(k)
            H_csr%j(k,ip)=positions(k)
          end do
! fill the lower half of the matrix:
! (if ntime=2, shift the columns by nlev, see the document)
          do k=H_csr%nrow+1,NUMB_NON_ZERO*H_csr%nrow
            H_csr%H(k,ip)=weights(k)
            H_csr%j(k,ip)=nlev*(ntime-1)+positions(k)
          end do
          do k=1,H_csr%nrow+1
            H_csr%i(k,ip)=(k-1)*NUMB_NON_ZERO+1
          end do  
        end if
      end do
      deallocate(weights)
      deallocate(positions)

      return
      end subroutine transf_vert_interp_matrix

!******************************************************************
!******************************************************************
! THE ROUTINE transf_temp_interp_matrix IS RESPONSIBLE FOR THE PREPARING 
! THE MATRIX FOR TEMPORAL INTERPOLATION OF THE MODEL DATA TO THE TIME OF 
! OBS USING THE TWO MOCAGE INTERVALS at T and T+DT.  
!******************************************************************
!******************************************************************
!******************************************************************
      subroutine transf_temp_interp_matrix(mlev,NPRF,ntime,date_in,date_out, &
                                           H_csr  )
!
      use floats 
      use MATRIX_CSR_STRUCTURE 
      use matrix_manipulations
!
      implicit none

! input variables:
      integer,intent(in)                       :: mlev   !obs levels
      integer,intent(in)                       :: ntime  !2 model times
      integer,intent(in)                       :: NPRF   !# of profiles
!
      real(std), dimension(ntime),intent(in)   ::date_in !the two dates of the model
      real(std), dimension(NPRF),intent(in)    ::date_out !the dates of the obs

! output variable
!  The vertical interp matrix in the CSR representation (structure with 3 ARRAYS):
      type(CSR_FORMAT),intent(out)                     :: H_csr  !REAL ARRAY


! local variables:
      real(std)                     :: fact
      real(std)                     :: w1,w2   !weights
      integer                       :: i,n,k
      integer                       :: nmax
      integer*4 :: ALLOC_ERR

      integer*4,parameter  :: NUMB_NON_ZERO=2  !number of non-zero elements


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! the code stars here:
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! construct the sparse matrix in the CRS format:

! define the dimensions of the CRS arrays
      H_csr%NPRF  = NPRF
      H_csr%nrow  = mlev
      H_csr%ncol  = mlev*ntime
      nmax        = NUMB_NON_ZERO*H_csr%nrow    !there are 2 non 0 elements in each row
      H_csr%nmax=nmax

      allocate(H_csr%H(nmax,NPRF),STAT = ALLOC_ERR)
      call check_allocate(ALLOC_ERR,"transf_temp_interp_matrix")
      allocate(H_csr%j(nmax,NPRF),STAT = ALLOC_ERR)
      call check_allocate(ALLOC_ERR,"transf_temp_interp_matrix")
      allocate(H_csr%i(H_csr%nrow+1,NPRF),STAT = ALLOC_ERR)
      call check_allocate(ALLOC_ERR,"transf_temp_interp_matrix")

      if (ntime.ne.2) then 
        write(*,*) 'ERROR, the time interpolation routine was called for ntime which is not 2'
        call exit(-1)
      end if
      fact=1./(date_in(2)-date_in(1))
!!$      WRITE(14871,*) 'Inner Ht_csr dates'
!!$      WRITE(14871,*) date_in(1), date_in(2)
!!$      WRITE(14871,*) fact
!!$      WRITE(*,*) 'Inner Ht_csr dates'
!!$      WRITE(*,*) date_in(1), date_in(2)
!!$      WRITE(*,*) fact
      do i=1,NPRF
        w1=(date_in(2)-date_out(i))*fact
        w2=(date_out(i)-date_in(1))*fact
!!$	WRITE(14871,*) 'Inner Ht_csr for i', i
!!$	WRITE(14871,*) w1
!!$	WRITE(14871,*) w2
!!$	
!!$	WRITE(*,*) 'Inner Ht_csr for i', i
!!$	WRITE(*,*) w1
!!$	WRITE(*,*) w2
! create matrix: 
!   first create the real array with non zero values: (see the corresponding document)
        do k=1,mlev
          H_csr%H((k-1)*2+1,i)=w1
          H_csr%H((k-1)*2+2,i)=w2
        end do
!   next the integer array with the column position of the NZ elements of the matrix
        do k=1,mlev
          H_csr%j((k-1)*2+1,i)=k
          H_csr%j((k-1)*2+2,i)=k+mlev
        end do
!   and finally the integer array with the pointers to the first element in 
!   each row of the full matrix.
        do k=1,mlev+1
          H_csr%i(k,i)=(k-1)*NUMB_NON_ZERO+1
        end do
      end do
!
      return
      end subroutine transf_temp_interp_matrix

!******************************************************************
!******************************************************************
!******************************************************************

      end module transformations_matrix









