MODULE SPECTRAL_FIELDS

!   Author.
!   -------
!     Yannick Tremolet
!
!   Modifications.
!   -------------- 
!     Original    02-03-05  
!     Y.Tremolet  02-07-25 Added operators.
!     C.Fischer   03-03-03 Aladin spectral treatment added.
!      N.B.1: the mean wind spec%sp1d is treated in assign_sp_sp and in
!             assign_scalar_sp but not in assign_ar_sp and assign_sp_ar.
!             therefore, whenever a fortran array is copied to a spectral_field
!             type in Aladin, one should take care of the mean wind externally.
!             see for example get_cv_init and set_cv_init in control_vectors.F90.
!     Y.Tremolet  11-Mar-2004 Type definition moved to SPECTRAL_FIELDS_DATA.
!     Y.Tremolet  11-Mar-2004 SPECTRAL_FIELDS_PARA for distributed operators.
! ------------------------------------------------------------------

USE PARKIND1, ONLY: JPIM, JPRB
USE SPECTRAL_FIELDS_DATA, ONLY: SPECTRAL_FIELD, &
                              & NPRTRV, NPRTRW, MYSETV, MYSETW, NFLEVL, &
                              & NFLEVG, NALLMS, NPTRMS, NUMLL, NPTRLL, &
                              & LELAM, NPSURF, MYMS, MYLEVS, LSPRT, NULOUT, &
                              & NGRBVO,NGRBD,NGRBT,NGRBQ,NGRBO3,NGRBLNSP
USE SPECTRAL_FIELDS_PARA, ONLY: SPECTRAL_NORM, SPECTRAL_NORM_LEVS, DOT_PRODUCT

IMPLICIT NONE
PRIVATE
PUBLIC SPECTRAL_FIELD, SETUP_SPEC, ALLOCATE_SPEC, DEALLOCATE_SPEC, &
     & ASSIGNMENT(=), OPERATOR(.EQV.), OPERATOR(.NEQV.), DOT_PRODUCT, &
     & SPECTRAL_NORM, SPECTRAL_NORM_LEVS, INQ_SPEC

! ------------------------------------------------------------------

INTERFACE ASSIGNMENT (=)
MODULE PROCEDURE assign_sp_sp, assign_scalar_sp, assign_ar_sp, assign_sp_ar
END INTERFACE

INTERFACE OPERATOR (.EQV.)
MODULE PROCEDURE equiv_spec
END INTERFACE

INTERFACE OPERATOR (.NEQV.)
MODULE PROCEDURE nequiv_spec
END INTERFACE

! ------------------------------------------------------------------
CONTAINS
! ------------------------------------------------------------------

SUBROUTINE setup_spec(kprtrv,kprtrw,ksetv,ksetw,kflevl,kflevg,knms,kmyms, &
                    & ksmax,kmsmax,kallms,kptrms,knmull,kptrll,kpsurf,ldlam, &
                    & ldsprt,kulout,kgrbvo,kgrbd,kgrbt,kgrbq,kgrbo3,kgrblnsp)

! This is a way to get values from ifs to ifsaux

INTEGER(KIND=JPIM), INTENT(IN) :: kprtrv,kprtrw,ksetv,ksetw,kflevl,kflevg,knms, &
                       & kmyms(knms),ksmax, kallms(ksmax+1),kptrms(kprtrw), &
                       & knmull(kprtrv),kptrll(kprtrv),kpsurf(kprtrv), &
                       & kulout,kgrbvo,kgrbd,kgrbt,kgrbq,kgrbo3,kgrblnsp,kmsmax
LOGICAL, INTENT(IN) :: ldlam, LDSPRT

INTEGER(KIND=JPIM) :: jl, iiismax

LELAM=ldlam

if (ldlam) then
  iiismax=kmsmax
else
  iiismax=ksmax
endif

NPRTRV=kprtrv
NPRTRW=kprtrw
MYSETV=ksetv
MYSETW=ksetw
NFLEVL=kflevl
NFLEVG=kflevg
IF (.NOT.ALLOCATED(MYMS))   ALLOCATE(MYMS(knms))
MYMS(:)=kmyms(:)
IF (.NOT.ALLOCATED(NALLMS)) ALLOCATE(NALLMS(iiismax+1))
NALLMS(:)=kallms(:)
IF (.NOT.ALLOCATED(NPTRMS)) ALLOCATE(NPTRMS(kprtrw))
NPTRMS(:)=kptrms(:)
IF (.NOT.ALLOCATED(NUMLL))  ALLOCATE(NUMLL(kprtrv))
NUMLL(:)=knmull(:)
IF (.NOT.ALLOCATED(NPTRLL)) ALLOCATE(NPTRLL(kprtrv))
NPTRLL(:)=kptrll(:)
IF (.NOT.ALLOCATED(NPSURF)) ALLOCATE(NPSURF(kprtrv))
NPSURF(:)=kpsurf(:)
IF (.NOT.ALLOCATED(MYLEVS)) ALLOCATE(MYLEVS(NFLEVL))
DO jl=1,NFLEVL
  MYLEVS(jl) = NPTRLL(MYSETV)+jl-1
ENDDO
LSPRT=LDSPRT
NULOUT=kulout
NGRBVO=kgrbvo
NGRBD =kgrbd
NGRBT =kgrbt
NGRBQ =kgrbq
NGRBO3=kgrbo3
NGRBLNSP=kgrblnsp

write(NULOUT,*)'SETUP_SPEC: SPECTRAL_FIELD type has been setup.'

RETURN
END SUBROUTINE setup_spec

! ------------------------------------------------------------------

SUBROUTINE allocate_spec(spec,ksmax,kmsmax,ks3d,ks2d,ks1d,kgrib)

TYPE (SPECTRAL_FIELD), INTENT(OUT) :: spec
INTEGER(KIND=JPIM), INTENT(IN) :: ksmax, kmsmax, ks3d, ks2d, ks1d, kgrib(ks3d+ks2d)

INTEGER(KIND=JPIM) :: idum1,idum2(0:ksmax),idum3,idum4, &
           & idum5(NPRTRW+1),idum6(0:ksmax),jp,jm,imaxms, &
           & imdim,il,ik,ind,ipos,ikm
INTEGER(KIND=JPIM) :: indd, indv, indt, indq, indo, inds
INTEGER(KIND=JPIM) :: IKNTMP(0:kmsmax),IKMTMP(0:ksmax),icpl4m(0:kmsmax)

IF (LELAM) THEN
  CALL ELLIPS(ksmax,kmsmax,IKNTMP,IKMTMP)
  imaxms=kmsmax/NPRTRW+1
  ALLOCATE(spec%nesm0 (0:kmsmax))
  ALLOCATE(spec%nesm0g(0:kmsmax))
  ALLOCATE(spec%ncpl4m(0:kmsmax))
  IF (ASSOCIATED(spec%nasm0 )) NULLIFY(spec%nasm0 )
  IF (ASSOCIATED(spec%nasm0g)) NULLIFY(spec%nasm0g)
ELSE
  imaxms=ksmax/NPRTRW+1
  ALLOCATE(spec%nasm0 (0:ksmax))
  ALLOCATE(spec%nasm0g(0:ksmax))
  NULLIFY(spec%nesm0 )
  NULLIFY(spec%nesm0g)
  NULLIFY(spec%ncpl4m)
ENDIF

ALLOCATE(spec%numpp(NPRTRW))
ALLOCATE(spec%ms_proc(imaxms,NPRTRW))
ALLOCATE(spec%numll(NPRTRV))
ALLOCATE(spec%nptrll(NPRTRV))
ALLOCATE(spec%ns2l(NPRTRV))
ALLOCATE(spec%ngrib(ks3d+ks2d))

spec%nsmax=ksmax
spec%ns2g=ks2d
spec%ns2l(:)=ks2d*NPSURF(:)
spec%ns2d=spec%ns2l(MYSETV)
spec%ns3d=ks3d

spec%nflevl=nflevl
spec%nflevg=nflevg

IF (LELAM) THEN
  spec%ns1d=ks1d
  spec%nmsmax=kmsmax
  spec%nspec2g=0
  do jm=0,kmsmax
    icpl4m(jm) = 4*(IKNTMP(jm)+1)
    spec%nspec2g = spec%nspec2g + 4*(IKNTMP(jm)+1)
  enddo
  spec%ncpl4m(:)=icpl4m(:)
  spec%nesm0g(0)=1
  do jm=1,kmsmax
    spec%nesm0g(jm) = spec%nesm0g(jm-1) + icpl4m(jm-1)
  enddo

  spec%numpp(:) = 0
  spec%nesm0(:) = -99

  imdim = 0
  il = 1
  ind = 1
  ik  = 0
  ipos = 1
  do jm=0,kmsmax
    ik = ik + ind
    if (ik > NPRTRW) then
      ik = NPRTRW
      ind = -1
    elseif (ik < 1) THEN
      ik = 1
      ind = 1
    endif

    ikm = icpl4m(jm)/4 -1
    spec%numpp(ik) = spec%numpp(ik) +1
    if (ik == MYSETW) THEN
      imdim = imdim + ikm +1
      spec%nesm0(jm) = ipos
      ipos = ipos + (ikm+1)*4
      il = il +1
    endif
  enddo

  spec%nspec2 = 4*imdim
ELSE
  spec%ns1d=0
  spec%nmsmax=0
  spec%nspec2g=(ksmax+1)*(ksmax+2)
  spec%nasm0g(0)=1
  DO jm=0,ksmax-1
    spec%nasm0g(jm+1) = spec%nasm0g(jm) + 2*(ksmax-jm+1)
  ENDDO

  CALL SUWAVEDI(ksmax,ksmax,NPRTRW,MYSETW, &
            & spec%nasm0,idum1,idum2,spec%numpp, &
            & idum3,spec%nspec2,idum4,idum5,idum6)
ENDIF

spec%nump = spec%numpp(MYSETW)
DO jp=1,NPRTRW
  DO jm=1,spec%numpp(jp)
    spec%ms_proc(jm,jp)=NALLMS(NPTRMS(jp)-1+jm)
  ENDDO
ENDDO

spec%numll(:) =NUMLL(:)
spec%nptrll(:)=NPTRLL(:)

spec%ngrib(:)=kgrib(:)

ALLOCATE(spec%sp2d(spec%nspec2,spec%ns2d))
ALLOCATE(spec%sp3d(spec%nflevl,spec%nspec2,spec%ns3d))
IF (LELAM) THEN
  ALLOCATE(spec%sp1d(spec%nflevl,spec%ns1d))
ELSE
  NULLIFY(spec%sp1d)
ENDIF

!cf nspsizeg and nspsizel contain only the horizontal spectral fields
! (eg. not the mean wind profiles for aladin case)

spec%nspsizel = (spec%nflevl*spec%ns3d+spec%ns2d)*spec%nspec2
spec%nspsizeg = (spec%nflevg*spec%ns3d+spec%ns2g)*spec%nspec2g

NULLIFY(spec%DIV)
NULLIFY(spec%VOR)
NULLIFY(spec%TEM)
NULLIFY(spec%HUM)
NULLIFY(spec%OZN)
NULLIFY(spec%SPR)
indd=IFIND(spec%ns3d,spec%ngrib,NGRBD)
indv=IFIND(spec%ns3d,spec%ngrib,NGRBVO)
indt=IFIND(spec%ns3d,spec%ngrib,NGRBT)
indq=IFIND(spec%ns3d,spec%ngrib,NGRBQ)
indo=IFIND(spec%ns3d,spec%ngrib,NGRBO3)
inds=IFIND(spec%ns2d,spec%ngrib(ks3d+1:ks3d+ks2d),NGRBLNSP)
IF (indd>0) spec%DIV => spec%sp3d(:,:,indd)
IF (indv>0) spec%VOR => spec%sp3d(:,:,indv)
IF (indt>0) spec%TEM => spec%sp3d(:,:,indt)
IF (indq>0) spec%HUM => spec%sp3d(:,:,indq)
IF (indo>0) spec%OZN => spec%sp3d(:,:,indo)
IF (inds>0) spec%SPR => spec%sp2d(:,inds)

spec%lvirtem=LSPRT

RETURN

CONTAINS
  INTEGER(KIND=JPIM) FUNCTION IFIND(kdim,kindx,kval)
  INTEGER(KIND=JPIM), INTENT(IN) :: kdim,kindx(:),kval
  INTEGER(KIND=JPIM) :: ii,jj
  jj=0
  ii=1
  DO WHILE (ii<=kdim .AND. jj==0)
    IF (kindx(ii)==kval) jj=ii
    ii=ii+1
  ENDDO
  ifind=jj
  END FUNCTION
END SUBROUTINE allocate_spec

! ------------------------------------------------------------------

SUBROUTINE deallocate_spec(spec)

TYPE (SPECTRAL_FIELD), INTENT(INOUT) :: spec

NULLIFY(spec%div)
NULLIFY(spec%vor)
NULLIFY(spec%tem)
NULLIFY(spec%hum)
NULLIFY(spec%ozn)
NULLIFY(spec%spr)

IF (ASSOCIATED(spec%sp2d   )) DEALLOCATE(spec%sp2d)
IF (ASSOCIATED(spec%sp3d   )) DEALLOCATE(spec%sp3d)
IF (ASSOCIATED(spec%sp1d   )) DEALLOCATE(spec%sp1d)
IF (ASSOCIATED(spec%nasm0  )) DEALLOCATE(spec%nasm0)
IF (ASSOCIATED(spec%nasm0g )) DEALLOCATE(spec%nasm0g)
IF (ASSOCIATED(spec%nesm0  )) DEALLOCATE(spec%nesm0)
IF (ASSOCIATED(spec%nesm0g )) DEALLOCATE(spec%nesm0g)
IF (ASSOCIATED(spec%ncpl4m )) DEALLOCATE(spec%ncpl4m)
IF (ASSOCIATED(spec%numpp  )) DEALLOCATE(spec%numpp)
IF (ASSOCIATED(spec%ms_proc)) DEALLOCATE(spec%ms_proc)
IF (ASSOCIATED(spec%numll  )) DEALLOCATE(spec%numll)
IF (ASSOCIATED(spec%nptrll )) DEALLOCATE(spec%nptrll)
IF (ASSOCIATED(spec%ns2l   )) DEALLOCATE(spec%ns2l)
IF (ASSOCIATED(spec%ngrib  )) DEALLOCATE(spec%ngrib)

RETURN
END SUBROUTINE deallocate_spec

! ------------------------------------------------------------------

SUBROUTINE assign_scalar_sp(spec,scalar)
TYPE (SPECTRAL_FIELD), INTENT(INOUT) :: spec
REAL(KIND=JPRB), INTENT(IN) :: scalar

spec%sp2d(:,:)  =scalar
spec%sp3d(:,:,:)=scalar
IF (LELAM) spec%sp1d(:,:)  =scalar

RETURN
END SUBROUTINE assign_scalar_sp

! ------------------------------------------------------------------

SUBROUTINE assign_sp_sp(spec1,spec2)
TYPE (SPECTRAL_FIELD), INTENT(INOUT) :: spec1
TYPE (SPECTRAL_FIELD), INTENT(IN)    :: spec2
INTEGER(KIND=JPIM) :: jf,jm,im,len,js1,js2

IF (spec1.NEQV.spec2) CALL ABOR1('Incompatible spectral fields')

IF ((spec1%nsmax<spec2%nsmax).AND.(.NOT.LELAM)) THEN      ! Truncate in IFS
  DO jf=1,spec1%ns2d
    DO jm=1,spec1%NUMP
      im=MYMS(jm)   
      len=2*(spec1%NSMAX-im+1)-1
      js1=spec1%NASM0(im)
      js2=spec2%NASM0(im)
      spec1%sp2d(js1:js1+len,jf) = spec2%sp2d(js2:js2+len,jf)
    ENDDO
  ENDDO
  DO jf=1,spec1%ns3d
    DO jm=1,spec1%NUMP
      im=MYMS(jm)   
      len=2*(spec1%NSMAX-im+1)-1
      js1=spec1%NASM0(im)
      js2=spec2%NASM0(im)
      spec1%sp3d(:,js1:js1+len,jf) = spec2%sp3d(:,js2:js2+len,jf)
    ENDDO
  ENDDO 
ELSEIF ((spec1%nsmax>spec2%nsmax).AND.(.NOT.LELAM)) THEN  ! Pad with zeros in IFS
  spec1%sp2d(:,:) = 0.0_JPRB
  DO jf=1,spec2%ns2d
    DO jm=1,spec2%NUMP
      im=MYMS(jm)   
      len=2*(spec2%NSMAX-im+1)-1
      js1=spec1%NASM0(im)
      js2=spec2%NASM0(im)
      spec1%sp2d(js1:js1+len,jf) = spec2%sp2d(js2:js2+len,jf)
    ENDDO
  ENDDO
  spec1%sp3d(:,:,:) = 0.0_JPRB
  DO jf=1,spec2%ns3d
    DO jm=1,spec2%NUMP
      im=MYMS(jm)   
      len=2*(spec2%NSMAX-im+1)-1
      js1=spec1%NASM0(im)
      js2=spec2%NASM0(im)
      spec1%sp3d(:,js1:js1+len,jf) = spec2%sp3d(:,js2:js2+len,jf)
    ENDDO
  ENDDO 
ELSEIF ((((spec1%nsmax<spec2%nsmax).AND.(spec1%nmsmax<=spec2%nmsmax)).OR. &
 &((spec1%nsmax<=spec2%nsmax).AND.(spec1%nmsmax<spec2%nmsmax))) &
 &.AND.(LELAM)) THEN      ! Truncate in Aladin
  DO jf=1,spec1%ns2d
    DO jm=1,spec1%NUMP
      im=MYMS(jm)   
      len=spec1%ncpl4m(im)-1
      js1=spec1%NESM0(im)
      js2=spec2%NESM0(im)
      spec1%sp2d(js1:js1+len,jf) = spec2%sp2d(js2:js2+len,jf)
    ENDDO
  ENDDO
  DO jf=1,spec1%ns3d
    DO jm=1,spec1%NUMP
      im=MYMS(jm)   
      len=spec1%ncpl4m(im)-1
      js1=spec1%NESM0(im)
      js2=spec2%NESM0(im)
      spec1%sp3d(:,js1:js1+len,jf) = spec2%sp3d(:,js2:js2+len,jf)
    ENDDO
  ENDDO 
ELSEIF ((((spec1%nsmax>spec2%nsmax).AND.(spec1%nmsmax>=spec2%nmsmax)).OR. &
 &((spec1%nsmax>=spec2%nsmax).AND.(spec1%nmsmax>spec2%nmsmax))) &
 &.AND.(LELAM)) THEN  ! Pad with zeros in Aladin
  spec1%sp2d(:,:) = 0.0_JPRB
  DO jf=1,spec2%ns2d
    DO jm=1,spec2%NUMP
      im=MYMS(jm)   
      len=spec2%ncpl4m(im)-1
      js1=spec1%NESM0(im)
      js2=spec2%NESM0(im)
      spec1%sp2d(js1:js1+len,jf) = spec2%sp2d(js2:js2+len,jf)
    ENDDO
  ENDDO
  spec1%sp3d(:,:,:) = 0.0_JPRB
  DO jf=1,spec2%ns3d
    DO jm=1,spec2%NUMP
      im=MYMS(jm)   
      len=spec2%ncpl4m(im)-1
      js1=spec1%NESM0(im)
      js2=spec2%NESM0(im)
      spec1%sp3d(:,js1:js1+len,jf) = spec2%sp3d(:,js2:js2+len,jf)
    ENDDO
  ENDDO 
ELSEIF (((spec1%nsmax/=spec2%nsmax).OR.(spec1%nmsmax/=spec2%nmsmax)).AND. &
 (LELAM)) THEN  ! Unfeasible cases left in Aladin
  CALL ABOR1('ASSIGN_SP_SP error: mismatch in Aladin spectral truncations')
ELSE                                   ! Copy
  spec1%sp2d(:,:)  =spec2%sp2d(:,:)
  spec1%sp3d(:,:,:)=spec2%sp3d(:,:,:)
ENDIF

! mean wind always copied if present in input
DO jf=1,spec2%ns1d
  spec1%sp1d(:,jf)=spec2%sp1d(:,jf)
ENDDO

RETURN
END SUBROUTINE assign_sp_sp

! ------------------------------------------------------------------

SUBROUTINE assign_ar_sp(spec,pflat)
TYPE (SPECTRAL_FIELD), INTENT(INOUT) :: spec
REAL(KIND=JPRB), INTENT(IN) :: pflat(:)
INTEGER(KIND=JPIM) :: i2d,i3d,ishape2d(2),ishape3d(3)
INTEGER(KIND=JPIM) :: ii,im,ioff,il,jsg,jf,jm,js,jn,jir,jl,len

IF (SIZE(pflat)==spec%nspsizel) THEN             ! local array (IFS and Aladin)

  i2d=spec%ns2d*spec%nspec2
  i3d=spec%nflevl*spec%ns3d*spec%nspec2
  ishape2d(1)=spec%nspec2
  ishape2d(2)=spec%ns2d
  ishape3d(1)=spec%nflevl
  ishape3d(2)=spec%nspec2
  ishape3d(3)=spec%ns3d
  spec%sp2d(:,:)  =RESHAPE(pflat(    1:    i2d),ishape2d)
  spec%sp3d(:,:,:)=RESHAPE(pflat(i2d+1:i2d+i3d),ishape3d)

ELSEIF ((SIZE(pflat)==spec%nspsizeg).AND.(.NOT.LELAM)) THEN         ! global array in IFS

  ioff=0
  DO jf=1,spec%ns2d
    DO jm=1,spec%NUMP
      im=MYMS(jm)
      jsg=spec%NASM0G(im)-1
      js =spec%NASM0 (im)-1
      len=2*(spec%nsmax-im+1)
      spec%sp2d(js+1:js+len,jf)=pflat(ioff+jsg+1:ioff+jsg+len)
    ENDDO
    ioff=ioff+spec%nspec2g
  ENDDO
  ioff=spec%ns2g*spec%nspec2g
  DO jf=1,spec%ns3d
    DO jm=1,spec%NUMP
      im=MYMS(jm)
      DO jn=im,spec%nsmax
        DO jir=0,1
          jsg=spec%NASM0G(im)+(jn-im)*2+jir
          js =spec%NASM0 (im)+(jn-im)*2+jir
          ii=ioff+(jsg-1)*spec%nflevg
          DO jl=1,spec%nflevl
            il=MYLEVS(jl)
            spec%sp3d(jl,js,jf)=pflat(ii+il)
          ENDDO
        ENDDO
      ENDDO
    ENDDO
    ioff=ioff+spec%nspec2g*spec%nflevg
  ENDDO

ELSEIF ((SIZE(pflat)==spec%nspsizeg).AND.(LELAM)) THEN         ! global array in Aladin

  ioff=0
  DO jf=1,spec%ns2d
    DO jm=1,spec%NUMP
      im=MYMS(jm)
      jsg=spec%NESM0G(im)-1
      js =spec%NESM0 (im)-1
      len=spec%ncpl4m(im)
      spec%sp2d(js+1:js+len,jf)=pflat(ioff+jsg+1:ioff+jsg+len)
    ENDDO
    ioff=ioff+spec%nspec2g
  ENDDO
  ioff=spec%ns2g*spec%nspec2g
  DO jf=1,spec%ns3d
    DO jm=1,spec%NUMP
      im=MYMS(jm)
      DO jn=1,spec%ncpl4m(im)
        jsg=spec%NESM0G(im)+jn-1
        js =spec%NESM0 (im)+jn-1
        ii=ioff+(jsg-1)*spec%nflevg
        DO jl=1,spec%nflevl
          il=MYLEVS(jl)
          spec%sp3d(jl,js,jf)=pflat(ii+il)
        ENDDO
      ENDDO
    ENDDO
    ioff=ioff+spec%nspec2g*spec%nflevg
  ENDDO

ELSE
  CALL ABOR1('assign_ar_sp: cannot assign array to spectral field')
ENDIF

!cf mean wind must be copied apart, but for conversions from spectral_field
!cf to control_vector and back, 0. seems to be the most appropriate value

IF (LELAM) spec%sp1d(:,:)=0.0_JPRB

RETURN
END SUBROUTINE assign_ar_sp

! ------------------------------------------------------------------

SUBROUTINE assign_sp_ar(pflat,spec)
REAL(KIND=JPRB), INTENT(OUT) :: pflat(:)
TYPE (SPECTRAL_FIELD), INTENT(IN) :: spec
INTEGER(KIND=JPIM) :: i2d,i3d,ishape2d(1),ishape3d(1)

IF (SIZE(pflat)==spec%nspsizel) THEN         ! local array (IFS and Aladin)

  i2d=spec%ns2d*spec%nspec2
  i3d=spec%nflevl*spec%ns3d*spec%nspec2
  ishape2d(1)=i2d
  ishape3d(1)=i3d
  pflat(    1:    i2d)=RESHAPE(spec%sp2d(:,:)  ,ishape2d)
  pflat(i2d+1:i2d+i3d)=RESHAPE(spec%sp3d(:,:,:),ishape3d)

ELSEIF (SIZE(pflat)==spec%nspsizeg) THEN     ! global array
  CALL ABOR1('Cannot assign spectral field to global array. '// &
           & 'Please use ALLGATHER_SPEC')
ELSE
  CALL ABOR1('Vector wrong length')
ENDIF

!cf information from spec%sp1d is not retrieved here - external copy required

RETURN
END SUBROUTINE assign_sp_ar

! ------------------------------------------------------------------

LOGICAL FUNCTION equiv_spec(spec1,spec2)
TYPE(SPECTRAL_FIELD), INTENT(IN) :: spec1
TYPE(SPECTRAL_FIELD), INTENT(IN) :: spec2
LOGICAL :: ll
INTEGER(KIND=JPIM) :: jf

ll = .true.
ll = ll .AND. (spec1%ns2g ==spec2%ns2g)
ll = ll .AND. (spec1%ns3d ==spec2%ns3d)
ll = ll .AND. (spec1%ns1d ==spec2%ns1d)
IF (ll) THEN
  DO jf=1,spec1%ns2g+spec1%ns3d
    ll = ll .AND. (spec1%ngrib(jf)==spec2%ngrib(jf))
  ENDDO
ENDIF

equiv_spec=ll

END FUNCTION equiv_spec

! ------------------------------------------------------------------

LOGICAL FUNCTION nequiv_spec(spec1,spec2)
TYPE(SPECTRAL_FIELD), INTENT(IN) :: spec1
TYPE(SPECTRAL_FIELD), INTENT(IN) :: spec2

nequiv_spec = .NOT. equiv_spec(spec1,spec2)

END FUNCTION nequiv_spec

! ------------------------------------------------------------------

SUBROUTINE inq_spec(spec,ksmax,kmsmax,ks3d,ks2d,ks1d)

! Sets all indices as allocate_spec without allocating
! space for data.

TYPE (SPECTRAL_FIELD), INTENT(OUT) :: spec
INTEGER(KIND=JPIM), INTENT(IN) :: ksmax, kmsmax, ks3d, ks2d, ks1d

INTEGER(KIND=JPIM) :: idum1,idum2(0:ksmax),idum3,idum4, &
           & idum5(NPRTRW+1),idum6(0:ksmax),jp,jm,imaxms, &
           & imdim,il,ik,ind,ipos,ikm
INTEGER(KIND=JPIM) :: IKNTMP(0:kmsmax),IKMTMP(0:ksmax),icpl4m(0:kmsmax)

IF (LELAM) THEN
  CALL ELLIPS(ksmax,kmsmax,IKNTMP,IKMTMP)
  imaxms=kmsmax/NPRTRW+1
  ALLOCATE(spec%nesm0 (0:kmsmax))
  ALLOCATE(spec%nesm0g(0:kmsmax))
  ALLOCATE(spec%ncpl4m(0:kmsmax))
  IF (ASSOCIATED(spec%nasm0 )) NULLIFY(spec%nasm0 )
  IF (ASSOCIATED(spec%nasm0g)) NULLIFY(spec%nasm0g)
ELSE
  imaxms=ksmax/NPRTRW+1
  ALLOCATE(spec%nasm0 (0:ksmax))
  ALLOCATE(spec%nasm0g(0:ksmax))
  NULLIFY(spec%nesm0 )
  NULLIFY(spec%nesm0g)
  NULLIFY(spec%ncpl4m)
ENDIF

ALLOCATE(spec%numpp(NPRTRW))
ALLOCATE(spec%ms_proc(imaxms,NPRTRW))
ALLOCATE(spec%numll(NPRTRV))
ALLOCATE(spec%nptrll(NPRTRV))
ALLOCATE(spec%ns2l(NPRTRV))

spec%nsmax=ksmax
spec%ns2g=ks2d
spec%ns2l(:)=ks2d*NPSURF(:)
spec%ns2d=spec%ns2l(MYSETV)
spec%ns3d=ks3d

spec%nflevl=nflevl
spec%nflevg=nflevg

IF (LELAM) THEN
  spec%ns1d=ks1d
  spec%nmsmax=kmsmax
  spec%nspec2g=0
  do jm=0,kmsmax
    icpl4m(jm) = 4*(IKNTMP(jm)+1)
    spec%nspec2g = spec%nspec2g + 4*(IKNTMP(jm)+1)
  enddo
  spec%ncpl4m(:)=icpl4m(:)
  spec%nesm0g(0)=1
  do jm=1,kmsmax
    spec%nesm0g(jm) = spec%nesm0g(jm-1) + icpl4m(jm-1)
  enddo

  spec%numpp(:) = 0
  spec%nesm0(:) = -99

  imdim = 0
  il = 1
  ind = 1
  ik  = 0
  ipos = 1
  do jm=0,kmsmax
    ik = ik + ind
    if (ik > NPRTRW) then
      ik = NPRTRW
      ind = -1
    elseif (ik < 1) THEN
      ik = 1
      ind = 1
    endif

    ikm = icpl4m(jm)/4 -1
    spec%numpp(ik) = spec%numpp(ik) +1
    if (ik == MYSETW) THEN
      imdim = imdim + ikm +1
      spec%nesm0(jm) = ipos
      ipos = ipos + (ikm+1)*4
      il = il +1
    endif
  enddo

  spec%nspec2 = 4*imdim
ELSE
  spec%ns1d=0
  spec%nmsmax=0
  spec%nspec2g=(ksmax+1)*(ksmax+2)
  spec%nasm0g(0)=1
  DO jm=0,ksmax-1
    spec%nasm0g(jm+1) = spec%nasm0g(jm) + 2*(ksmax-jm+1)
  ENDDO

  CALL SUWAVEDI(ksmax,ksmax,NPRTRW,MYSETW, &
            & spec%nasm0,idum1,idum2,spec%numpp, &
            & idum3,spec%nspec2,idum4,idum5,idum6)
ENDIF

spec%nump = spec%numpp(MYSETW)
DO jp=1,NPRTRW
  DO jm=1,spec%numpp(jp)
    spec%ms_proc(jm,jp)=NALLMS(NPTRMS(jp)-1+jm)
  ENDDO
ENDDO

spec%numll(:) =NUMLL(:)
spec%nptrll(:)=NPTRLL(:)

!cf nspsizeg and nspsizel contain only the horizontal spectral fields
! (eg. not the mean wind profiles for aladin case)

spec%nspsizel = (spec%nflevl*spec%ns3d+spec%ns2d)*spec%nspec2
spec%nspsizeg = (spec%nflevg*spec%ns3d+spec%ns2g)*spec%nspec2g

ALLOCATE(spec%ngrib(1))
ALLOCATE(spec%sp2d(1,1))
ALLOCATE(spec%sp3d(1,1,1))
IF (LELAM) THEN
  ALLOCATE(spec%sp1d(1,1))
ELSE
  NULLIFY(spec%sp1d )
ENDIF

RETURN
END SUBROUTINE inq_spec

! ------------------------------------------------------------------

END MODULE SPECTRAL_FIELDS
